//Flare Lua by SMS Alfredo
local P_TeleportMove = P_MoveOrigin or P_TeleportMove

sfxinfo[freeslot("sfx_flshot")].caption = "Fire shoot"
sfxinfo[freeslot("sfx_flburn")].caption = "Burn"
sfxinfo[freeslot("sfx_flmove")].caption = "/"
sfxinfo[freeslot("sfx_flbrst")].caption = "Burst"
sfxinfo[freeslot("sfx_fljump")].caption = "Burst"
sfxinfo[freeslot("sfx_flhop")].caption = "Hop"
sfxinfo[freeslot("sfx_flthrw")].caption = "Flamethrower"

sfxinfo[freeslot("sfx_flstrg")].caption = "/"
sfxinfo[sfx_flstrg].flags = $|SF_X2AWAYSOUND
sfxinfo[freeslot("sfx_flland")].caption = "/"
sfxinfo[sfx_flland].flags = $|SF_X2AWAYSOUND

local FIRESPRITES = {
	[SPR2_STND] = SPR2_FIRE,
	//[SPR2_WAIT] = SPR2_FIRE,
	[SPR2_EDGE] = SPR2_FIRE,
	
	[SPR2_SPNG] = SPR2_TWIN,
	[SPR2_FALL] = SPR2_TWIN,
	
	[SPR2_WALK] = SPR2_TAL0,
	[SPR2_RUN_] = SPR2_TAL1,
}

//Can't move conditions
local function CantMove(player, noskin, nocarry)
	local mo = player.mo
	return not (mo and mo.valid and (mo.skin == "flarethetrex" or noskin))
	or P_PlayerInPain(player)
	or (player.panim == PA_PAIN and not (player.pflags & PF_JUMPED))
	or player.pflags & PF_SLIDING
	or (player.powers[pw_carry] and not nocarry)
	or player.spectator
	or player.playerstate
	or not mo.health
	or player.powers[pw_super] == 1
	or player.iseggrobo
	or player.isjettysyn
	or player.guard
	or player.actionstate
	or player.powers[pw_nocontrol]
	or (player.isxmomentum and mo.state == S_PLAY_FACEPLANT)
end

//Reset Flare
local function ResetFlare(player)
	player.flarereset = true
	player.flarelanding = 0
	player.flarespinfire = false
	player.flaredetransform = false
	player.flarebouncex = nil
	player.flarebouncey = nil
end
addHook("PlayerSpawn", ResetFlare)
for player in players.iterate ResetFlare(player) end

//Spawns objects in a circle
rawset(_G, "flarecircle", function(mo, mobjtype, amount, ns, zoffset, scale, keepmom, color, vertical, vertangle)
	if not amount then return end
	if vertangle == nil
		vertangle = mo.player and mo.player.drawangle or mo.angle
	end
	
	local i = 0	while i < amount
		local fa = i*(ANGLE_180/(amount/2)) + (vertangle or 0)
		local mobj = P_SpawnMobjFromMobj(mo,0,0,zoffset,mobjtype)
		if mobj and mobj.valid
			local height = mobj.height
			mobj.scale = scale
			
			if mobj.flags & MF_MISSILE
				mobj.target = mo
			end
			
			if mobjtype == MT_MINECARTSPARK or mobj.type == MT_GHOST
				mobj.fuse = TICRATE/2
			end
			
			if mo.eflags&MFE_VERTICALFLIP
				mobj.z = $ - (mobj.height-height)
			end
			
			if color
				mobj.color = color
				if mobjtype != MT_GHOST
					mobj.colorized = true
				end
			end
			
			if vertical
				mobj.momz = FixedMul(sin(fa),ns)
				P_InstaThrust(mobj, vertangle+ANGLE_90,
				FixedMul(cos(fa),ns))
			else
				mobj.momx = FixedMul(cos(fa),ns)
				mobj.momy = FixedMul(sin(fa),ns)
				mobj.angle = fa
			end
			
			if keepmom == -1
				mobj.momx = $ + mo.momx*2/3
				mobj.momy = $ + mo.momy*2/3
				mobj.momz = $ + P_GetMobjGravity(mo)*8
			elseif keepmom
				if keepmom == true then keepmom = 1 end
				mobj.momx = $ + mo.momx/keepmom
				mobj.momy = $ + mo.momy/keepmom
				mobj.momz = $ + mo.momz/keepmom
			end
			
			if mobj.type == MT_GHOST
				mobj.momx = $ + P_RandomRange(-24, 24)*mobj.scale/6
				mobj.momy = $ + P_RandomRange(-24, 24)*mobj.scale/6
				mobj.momz = $ + P_RandomRange(2, 24)*mobj.scale/3*P_MobjFlip(mobj)
				if not vertical P_SetObjectMomZ(mobj, 4*FRACUNIT, true) end
			end
		end
		i = $ + 1
	end
end)

//Fire projectile
local FLAMETHOWERHOLD = 7
local FIREBALLDELAY = TICRATE/2
local SUPERFIRE = SKINCOLOR_MINT

freeslot("MT_FLARE_FIRE",
"S_FLARE_FIRE", "S_FLARE_EXPLODE", "S_FLARE_BURST", "S_FLARE_RESULT", "S_FLARE_RESULT2",
"SPR_FLFR", "SPR_FLEX", "SPR_FLBR", "SPR_FLRS")

function A_FlareExplode(mobj, var1, var2)
	if mobj.type != MT_FLARE_FIRE then return end
	
	A_Scream(mobj)
	mobj.shadowscale = 0
	
	local fire = P_SpawnMobjFromMobj(mobj, 0,0,0, MT_SPINFIRE)
	if not (fire and fire.valid) then return end
	
	//Clip up fix
	if P_IsObjectOnGround(fire)
		P_KillMobj(fire)
		return
	end
	
	if mobj.target and mobj.target.player
	and mobj.target.player.powers[pw_super]
		fire.color = SUPERFIRE
		fire.colorized = true
	else
		fire.color = mobj.color
		fire.colorized = mobj.colorized
	end
		
	fire.blendmode = AST_ADD
	fire.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
	fire.state = S_FLARE_RESULT
	
	fire.target = mobj.target
	fire.flags = $ &~ MF_NOGRAVITY &~ MF_FIRE
	fire.scalespeed = FixedMul($, fire.scale)
	fire.scale = $/2
	fire.destscale = $*4
	fire.spritexscale = $*3/2
	fire.spriteyscale = $*2/3
	
	fire.momz = P_GetMobjGravity(mobj)*-16
	fire.shadowscale = FRACUNIT
end

function A_FlareFade(mobj, var1, var2)
	mobj.type = MT_GHOST
	mobj.fuse = mobj.tics
	mobj.destscale = 1
	mobj.flags = MF_NOCLIP | MF_NOCLIPHEIGHT | MF_NOCLIPTHING | MF_NOGRAVITY
end

states[S_FLARE_FIRE] = {SPR_FLFR, FF_ANIMATE, O+1, nil, O, 1, S_FLARE_FIRE}
states[S_FLARE_EXPLODE] = {SPR_FLEX, FF_ANIMATE, M+1, A_FlareExplode, M, 1, S_BIGMINE_BLAST5}
states[S_FLARE_BURST] = {SPR_FLBR, FF_ANIMATE, (F+1)*2, nil, F, 2, S_NULL}
states[S_FLARE_RESULT] = {SPR_FLRS, FF_ANIMATE, TICRATE*4, nil, H, 1, S_FLARE_RESULT2}
states[S_FLARE_RESULT2] = {SPR_FLRS, FF_ANIMATE, 16, A_FlareFade, H, 1, S_NULL}
mobjinfo[MT_FLARE_FIRE] = {
	spawnstate = S_FLARE_FIRE,
	deathstate = S_FLARE_EXPLODE,
	seesound = sfx_flshot,
	deathsound = sfx_flburn,
	spawnhealth = 1000,
	speed = 60*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 16*FRACUNIT,
	damage = 1,
	mass = DMG_FIRE,
	flags = MF_MISSILE|MF_NOGRAVITY,
}

addHook("MobjThinker", function(mobj)
	if not mobj.threshold
		mobj.blendmode = AST_ADD
		mobj.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
		mobj.spriteyscale = $*2
		mobj.spritexscale = $*2
		mobj.shadowscale = FRACUNIT
		
		if mobj.target and mobj.target.player
		and mobj.target.player.powers[pw_super]
			mobj.color = SUPERFIRE
			mobj.colorized = true
		end
		
		P_TryMove(mobj, mobj.x - mobj.momx, mobj.y - mobj.momy, true)
		mobj.z = $ - mobj.momz
		
		if not mobj.fuse
			mobj.fuse = TICRATE*5
		end
		mobj.threshold = 1
	end
	
	if mobj.fuse and not (mobj.flags & MF_MISSILE and mobj.health)
		mobj.fuse = 0
	end
	
	if not mobj.sprite then return end
	
	
	if mobj.flags & MF_MISSILE and mobj.health
		if not (mobj.tics % 4)
			local thok = P_SpawnMobjFromMobj(mobj, 0,0, mobj.info.height/2, MT_THOK)
			thok.color = SKINCOLOR_ICY
			thok.colorized = true
			thok.blendmode = AST_ADD
			
			thok.renderflags = $ | RF_PAPERSPRITE | RF_FULLBRIGHT | RF_NOCOLORMAPS
			thok.angle = mobj.angle + ANGLE_90
			
			thok.state = S_CACOFIRE_EXPLODE4
			thok.tics = 8
			thok.fuse = thok.tics
			
			thok.scalespeed = FixedMul($, thok.scale)
			thok.scale = $/4
			thok.destscale = $*64
			
			S_StartSound(thok, sfx_flmove)
		end
		
		P_RadiusAttack(mobj, mobj.target or mobj, mobj.info.radius*5, mobj.info.mass, true)
	end
	
	local ghost = P_SpawnGhostMobj(mobj)
	
	ghost.momx = P_SignedRandom()*mobj.scale/64
	ghost.momy = P_SignedRandom()*mobj.scale/64
	ghost.momz = P_SignedRandom()*mobj.scale/64
	
	ghost.blendmode = mobj.blendmode
	ghost.renderflags = mobj.renderflags
	ghost.spritexscale = mobj.spritexscale
	ghost.spriteyscale = mobj.spriteyscale
	
	ghost.frame = $ &~ FF_TRANSMASK
	ghost.fuse = TICRATE/2
end, MT_FLARE_FIRE)

local function firedie(lol, mobj)
	if mobj and mobj.type == MT_FLARE_FIRE then P_ExplodeMissile(mobj) end
end
addHook("MobjDamage", firedie)
addHook("MobjDeath", firedie)

//Flamethrower
freeslot("MT_FLARE_THROWER",
"S_FLARE_THROWER", "SPR_FLTH")
states[S_FLARE_THROWER] = {SPR_FLTH, FF_ANIMATE|FF_FLOORSPRITE, (D*2)+1, nil, D, 2, S_FLARE_THROWER}
mobjinfo[MT_FLARE_THROWER] = {
	spawnstate = S_FLARE_THROWER,
	deathstate = S_FLARE_THROWER,
	seesound = sfx_bexpld,
	spawnhealth = 1000,
	speed = FRACUNIT,
	radius = 96*FRACUNIT,
	height = 48*FRACUNIT,
	damage = 1,
	mass = DMG_FIRE,
	flags = MF_MISSILE|MF_NOGRAVITY|MF_BOUNCE|MF_GRENADEBOUNCE|MF_NOCLIPHEIGHT,
}

addHook("MobjThinker", function(mobj)
	if not mobj.threshold
		mobj.blendmode = AST_ADD
		mobj.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS | RF_NOSPLATBILLBOARD
		mobj.spritexscale = FRACUNIT/16
		
		mobj.flags = mobj.info.flags
		mobj.flags2 = $ | MF2_DONTDRAW
		mobj.rollangle = ANGLE_90
	end
	mobj.threshold = $ + 1
	mobj.movedir = mobj.angle
	A_ForceStop(mobj)
	A_CapeChase(mobj)
	
	if not (mobj and mobj.valid) then return end
	local mo = mobj.target
	local player = mo and mo.player
	if not player then return end
	
	if player.flarethrower != mobj
		player.flarethrower = mobj
	end
	
	if player.powers[pw_super]
		mobj.color = SUPERFIRE
		mobj.colorized = true
		mobj.scale = mo.scale*2
		
	elseif mobj.colorized
		mobj.colorized = false
		mobj.scale = mo.scale
	end
	
	if mobj.threshold <= 1 then mobj.movedir = mobj.angle end
	
	if mobj.threshold > 1 and (CantMove(player) or player.flarespin < FLAMETHOWERHOLD)
		mobj.spritexscale = $ - FRACUNIT/4
		if mobj.spritexscale <= FRACUNIT/16
			P_RemoveMobj(mobj)
			return
		end
	elseif mobj.spritexscale < FRACUNIT*2
		mobj.spritexscale = $ + FRACUNIT/4
	end
	
	if not S_SoundPlaying(mobj, sfx_flthrw)
		S_StartSound(mobj, sfx_flthrw)
	end
	
	local thrust = mo.radius + FixedMul(mobj.radius*2/3, mobj.spritexscale)
	P_TeleportMove(mobj, mobj.x + P_ReturnThrustX(mobj, mobj.angle, thrust),
	mobj.y + P_ReturnThrustY(mobj, mobj.angle, thrust), mobj.z)
	
	for i = 1, 4
		local ghost = P_SpawnGhostMobj(mobj)
		
		thrust = FixedMul(mobj.info.radius*2/3, mobj.spritexscale)/FRACUNIT
		thrust = P_RandomRange(-$, $)*mobj.scale
		P_TeleportMove(ghost, ghost.x + P_ReturnThrustX(mobj, mobj.movedir, thrust),
		ghost.y + P_ReturnThrustY(mobj, mobj.movedir, thrust), ghost.z + mobj.height/2*P_MobjFlip(mo))
		
		ghost.momx = P_SignedRandom()*mobj.scale/32 + mo.momx/2
		ghost.momy = P_SignedRandom()*mobj.scale/32 + mo.momy/2
		ghost.momz = P_SignedRandom()*mobj.scale/32 + mo.momz/2
		
		ghost.blendmode = mobj.blendmode
		ghost.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
		ghost.spritexscale = mobj.spritexscale
		ghost.spriteyscale = mobj.spriteyscale
		
		ghost.state = S_FLARE_EXPLODE
		ghost.fuse = 0
	end
end, MT_FLARE_THROWER)

//Burn rings
freeslot("MT_FLARE_BURNTRING")
mobjinfo[MT_FLARE_BURNTRING] = {
	spawnstate = S_TEAMRING,
	deathstate = S_SPRK1,
	painsound = sfx_itemup,
	spawnhealth = 1000,
	speed = 38*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	mass = 100,
	flags = MF_SLIDEME,
}

addHook("MobjThinker", function(mobj)
	if not mobj.health then return end
	
	A_BossScream(mobj, 0, MT_SMOKE)
	
	if P_IsObjectOnGround(mobj)
		P_KillMobj(mobj)
		A_ForceStop(mobj)
		
		mobj.shadowscale = 0
		mobj.colorized = true
		
		if mobj.target and mobj.target.player
			if mobj.target.player == displayplayer
			or mobj.target.player == secondarydisplayplayer
				S_StartSound(mobj.target, mobj.info.painsound)
			else
				S_StartSound(mobj, mobj.info.painsound)
			end
			
			P_GivePlayerRings(mobj.target.player, 1)
			if mapheaderinfo[gamemap].typeoflevel&TOL_NIGHTS
			and target.oldtype != MT_FLINGRING and target.oldtype != MT_FLINGCOIN
				P_DoNightsScore(mobj.target.player)
			end
		end
	end
end, MT_FLARE_BURNTRING)

local function burnrings(mobj, target)
	if target.player and gametyperules & GTR_FRIENDLY
	and not CV_FindVar("friendlyfire").value
		return false
	elseif (target.type == MT_RING or target.type == MT_FLINGRING
	or (target.type == MT_REDTEAMRING and mobj.target.player.ctfteam == 1)
	or (target.type == MT_BLUETEAMRING and mobj.target.player.ctfteam == 2))
	and mobj.health and mobj.threshold and mobj.target and mobj.target.player
	and P_CanPickupItem(mobj.target.player)
	and mobj.z - mobj.scale <= target.z + target.height
	and mobj.z + mobj.height + mobj.scale >= target.z
		P_SpawnGhostMobj(target)
			
		target.state = S_TEAMRING
		target.color = SKINCOLOR_CARBON
		
		target.oldtype = target.type
		target.type = MT_FLARE_BURNTRING
		target.flags = target.info.flags
		
		target.momx = mobj.momx/4 + P_RandomRange(-24, 24)*mobj.scale/6
		target.momy = mobj.momy/4 + P_RandomRange(-24, 24)*mobj.scale/6
		target.momz = $ + P_RandomRange(2, 24)*mobj.scale/2*P_MobjFlip(mobj)
		
		if mobj.target and mobj.target.player
		and (mobj.target.player == displayplayer
		or mobj.target.player == secondarydisplayplayer)
			S_StartSound(mobj.target, sfx_s3k4b)
		else
			S_StartSound(mobj, sfx_s3k4b)
		end
		target.target = mobj.target
		
		local thok = P_SpawnMobjFromMobj(target, 0,0,0, MT_THOK)
		
		thok.blendmode = mobj.blendmode
		thok.renderflags = mobj.renderflags
		thok.spritexscale = mobj.spritexscale
		thok.spriteyscale = mobj.spriteyscale
		
		thok.state = S_FLARE_EXPLODE
	end
end
addHook("MobjMoveCollide", burnrings, MT_FLARE_FIRE)
addHook("MobjMoveCollide", burnrings, MT_FLARE_THROWER)

//Spin abilities
addHook("SpinSpecial", function(player)
	local mo = player.mo
	if CantMove(player) then return end
	if player.pflags & PF_STASIS
	or player.exiting or P_SuperReady(player)
	or player.panim == PA_DASH
	then return true end
	
	if mo.state == S_PLAY_MELEE_LANDING
		if not (player.pflags & PF_SPINDOWN)
			player.pflags = $ | PF_SPINNING | PF_THOKKED &~ PF_JUMPED &~ PF_NOJUMPDAMAGE
			S_StartSound(mo, sfx_flbrst)
			player.glidetime = 1
			
			mo.state = S_PLAY_STND
			mo.state = S_PLAY_BOUNCE_LANDING
			
			P_InstaThrust(mo, mo.angle, 60*mo.scale)
			
			local thok = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THOK)
			thok.blendmode = AST_ADD
			thok.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
			thok.state = S_FLARE_BURST
			
			if player.powers[pw_super]
				thok.color = SUPERFIRE
				thok.colorized = true
				thok.scale = $*2
				
				P_Earthquake(mo, mo, 256*FRACUNIT)
			end
		end
	
		return true
	elseif (player.flarespin < FLAMETHOWERHOLD
	and player.pflags & PF_SPINDOWN) or player.weapondelay
	or (player.flarespin >= FLAMETHOWERHOLD and player.pflags & PF_SPINNING)
		return true
	end
	
	player.pflags = $ &~ PF_SPINNING
	if player.flarespinfire and not P_IsObjectOnGround(mo)
		player.pflags = $ | PF_JUMPED | PF_NOJUMPDAMAGE &~ PF_THOKKED
	elseif player.pflags & PF_JUMPED
		player.pflags = $ | PF_NOJUMPDAMAGE
	end
	
	player.powers[pw_justsprung] = max($, 5)
	if not FIRESPRITES[mo.sprite2 &~ FF_SPR2SUPER]
		mo.state = S_PLAY_FALL
	end
	
	if player.powers[pw_super]
	and player == displayplayer and not splitscreen
		P_StartQuake(player.flarespin >= FLAMETHOWERHOLD and 2*mo.scale or 8*mo.scale, 2)
	end
	
	if player.flarespin >= FLAMETHOWERHOLD
		if not (player.flarethrower and player.flarethrower.valid)
			player.flarethrower = P_SpawnPlayerMissile(mo, MT_FLARE_THROWER, 0)
			player.drawangle = mo.angle
			
			mo.momx = $/2
			mo.momy = $/2
			mo.momz = $/2
		
		//Slow turning
		elseif player.cmd.forwardmove or player.cmd.sidemove
			local curangle = player.drawangle
			local tarangle = mo.angle
			local factor = 16
			
			//Let's see how much we'll turn
			local diff = (tarangle - curangle)
			
			//Apply it!
			if not diff
				//
			elseif (diff > ANGLE_180)
				diff = InvAngle(InvAngle(diff)/factor)
			else
				diff = $/factor
			end
			
			player.drawangle = $ + diff
		end
		
		//Move slower
		if P_IsObjectOnGround(mo)
			mo.friction = $*7/8
		else
			mo.momz = $*15/16
			player.pflags = $ | PF_THOKKED
		end
	else
		player.drawangle = mo.angle
		player.weapondelay = FIREBALLDELAY
		P_SpawnPlayerMissile(mo, MT_FLARE_FIRE, 0)
		
		if player.powers[pw_super]
			S_StartSound(mo, sfx_flstrg)
			player.weapondelay = $*2
			
			P_SPMAngle(mo, MT_FLARE_FIRE, mo.angle + ANGLE_11hh)
			P_SPMAngle(mo, MT_FLARE_FIRE, mo.angle - ANGLE_11hh)
		end
	end
end)

//Jump Abilities
addHook("AbilitySpecial", function(player)
	local mo = player.mo
	if CantMove(player) then return end
	if player.pflags & PF_JUMPSTASIS or player.exiting
	then return true end
	
	//Already Thokked
	if player.pflags & PF_THOKKED
		//Super transformation/detransformation
		//Press jump during the upwards burst!
		if ((not player.powers[pw_super] and not player.powers[pw_invulnerability]
		and not (player.powers[pw_shield] & SH_NOSTACK) and player.rings >= 50)
		or player.powers[pw_super]) and player.glidetime == 2
		and not (player.pflags & PF_JUMPED) and player.pflags & PF_SPINNING
		and not (maptol & TOL_NIGHTS) and All7Emeralds(emeralds)
			P_ResetPlayer(player)
			if player.powers[pw_super]
				player.powers[pw_emeralds] = 0 // lost the power stones
				player.powers[pw_super] = 0
				P_SpawnGhostMobj(mo)
				P_RestoreMusic(player)
				P_SpawnShieldOrb(player)
				
				mo.state = S_PLAY_NIGHTS_TRANS1
				player.flaredetransform = true

				// Restore color
				if ((player.powers[pw_shield] & SH_STACK) == SH_FIREFLOWER)
					mo.color = SKINCOLOR_WHITE
				else
					mo.color = player.skincolor
				end

				// Inform the netgame that the champion has fallen in the heat of battle.
				if (gametyperules & (GTR_FRIENDLY|GTR_CAMPAIGN)) != (GTR_FRIENDLY|GTR_CAMPAIGN)
					player.powers[pw_flashing] = flashingtics-1
					S_StartSound(nil, sfx_s3k66) //let all players hear it.
				else
					S_StartSound(mo, sfx_s3k66)
				end
				
				for d = 0, 15
					P_SpawnParaloop(mo.x, mo.y, mo.z + mo.height, mo.radius, 16,
					MT_NIGHTSPARKLE, d*(ANGLE_22h), S_NULL, true)
				end
			else
				P_DoSuperTransformation(player, false)
			end
			
			return true
		
		//No multiability
		elseif not (player.charflags & SF_MULTIABILITY)
			return true
		end
	end
	
	S_StartSound(mo, sfx_s3k42)
	S_StartSound(mo, sfx_s3k48)
	mo.state = S_PLAY_GLIDE
	player.pflags = $ | PF_THOKKED
	player.flarelanding = 1
	
	return true
end)

addHook("JumpSpecial", function(player)
	local mo = player.mo
	if CantMove(player) or not (mo.state == S_PLAY_MELEE_LANDING
	or (P_IsObjectOnGround(mo) and (player.flarespinfire or player.pflags & PF_STARTDASH)))
	or player.pflags & (PF_JUMPSTASIS|PF_JUMPDOWN) or player.exiting
	then return end
	
	if player.flarespinfire or player.pflags & PF_STARTDASH
		P_DoJump(player, false)
		mo.momz = $*2/3
		
		local thrust = min(60*mo.scale - FixedHypot(mo.momx, mo.momy), 8*mo.scale)
		local angle = player.drawangle
		if player.cmd.forwardmove or player.cmd.sidemove
			angle = player.cmd.angleturn<<16 + R_PointToAngle2(0, 0, player.cmd.forwardmove*FRACUNIT, -player.cmd.sidemove*FRACUNIT)
		end
		P_Thrust(mo, angle, thrust)
		
		player.pflags = $ | PF_SPINNING | PF_THOKKED &~ PF_JUMPED &~ PF_NOJUMPDAMAGE &~ PF_STARTJUMP
		S_StartSound(mo, sfx_flhop)
		
		mo.state = S_PLAY_ROLL
		flarecircle(mo, mo.eflags & MFE_UNDERWATER and MT_SMALLBUBBLE or MT_DUST,
		8, 6*mo.scale, 0, mo.scale*2/3, 2)
		
		return true
	end
	
	player.pflags = $ &~ PF_JUMPED
	P_DoJump(player, false)
	mo.momz = $*2
	
	player.pflags = $ | PF_SPINNING | PF_THOKKED &~ PF_JUMPED &~ PF_NOJUMPDAMAGE
	S_StartSound(mo, sfx_fljump)
	player.glidetime = 2
	
	mo.state = S_PLAY_STND
	mo.state = S_PLAY_BOUNCE_LANDING
	player.panim = PA_DASH
	
	local thok = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THOK)
	thok.blendmode = AST_ADD
	thok.state = S_FLARE_BURST
	
	if player.powers[pw_super]
		thok.color = SUPERFIRE
		thok.colorized = true
		thok.scale = $*2
		
		P_Earthquake(mo, mo, 256*FRACUNIT)
	end
	
	return true
end)

//Jump ability bonk
addHook("MobjMoveBlocked", function(mo, mobj, line)
	local player = mo.player
	if not player or CantMove(player) or not (player.flarelanding
	or (player.pflags & PF_SPINNING and player.pflags & PF_THOKKED))
	or player.speed < 4*mo.scale or player.powers[pw_justlaunched]
	then return end
	
	local oldmomx = mo.momx
	local oldmomy = mo.momy
	
	P_BounceMove(mo)
	
	player.flarebouncex = mo.momx
	player.flarebouncey = mo.momy
	
	mo.momx = oldmomx
	mo.momy = oldmomy
end)

//Landing and bursts
local function flareland(player)
	if player.mo and player.mo.skin == "flarethetrex"
	and (player.mo.state == S_PLAY_MELEE_LANDING
	or player.mo.state == S_PLAY_BOUNCE_LANDING)
		return P_GetPlayerSpinHeight(player)
	end
end
addHook("PlayerHeight", flareland)
addHook("PlayerCanEnterSpinGaps", flareland)

//Main thinker
addHook("PlayerThink", function(player)
	local mo = player.mo
	
	if mo.skin == "flarethetrex"
		if player.flarespin == nil
			player.flarespin = 0
		elseif player.cmd.buttons & BT_SPIN and not player.flarespinfire
			player.flarespin = $ + 1
		elseif player.flarespin
			player.flarespin = 0
		end
		
		//Prevent transforming normally
		if not player.powers[pw_super] and not player.powers[pw_invulnerability]
		and not player.powers[pw_tailsfly] and (player.pflags & (PF_JUMPED|PF_THOKKED))
		and not (player.powers[pw_shield] & SH_NOSTACK)
		and not (maptol & TOL_NIGHTS) and All7Emeralds(emeralds)
		and player.rings >= 50 and not player.climbing and not player.exiting
		and not player.powers[pw_carry] and not P_PlayerInPain(player)
		and not (player.pflags & (PF_GLIDING|PF_SLIDING|PF_SHIELDABILITY|PF_JUMPSTASIS))
			if player.charflags & SF_SUPER
				player.charflags = $ &~ SF_SUPER
			end
		elseif not (player.charflags & SF_SUPER)
			player.charflags = $ | SF_SUPER
		end
		
		if not mo.flareshadow
			mo.flareshadow = true
			mo.shadowscale = FRACUNIT*5/4
		end
		
		if mo.flarecarrying
			mo.flarecarrying = $ - 1
		end
	elseif mo.flareshadow
		mo.flareshadow = false
		mo.shadowscale = FRACUNIT
	end
	
	//Prevent the rest of the code
	if CantMove(player)
		if not player.flarereset
			ResetFlare(player)
		end
		return
	end
	player.flarereset = false
	
	//Detransformation
	if player.flaredetransform
		if mo.state >= S_PLAY_NIGHTS_TRANS1 and mo.state <= S_PLAY_NIGHTS_TRANS6
			A_ForceStop(mo)
			player.pflags = $ | PF_FULLSTASIS
			return
		else
			player.flaredetransform = false
		end
	end
	
	//Makes it so that you properly bounce off of walls
	if player.flarebouncex != nil and player.flarebouncey != nil
		if not player.flarelanding
			player.pflags = $ | PF_SPINNING | PF_THOKKED
			if mo.state != S_PLAY_ROLL
				mo.state = S_PLAY_ROLL
			end
		end
		if not player.powers[pw_justlaunched]
			mo.momx = player.flarebouncex
			mo.momy = player.flarebouncey
			
			player.drawangle = R_PointToAngle2(mo.momx, mo.momy, 0,0)
			player.powers[pw_justsprung] = max($, 15)
			
			S_StartSound(mo, sfx_s3k5d)
			flarecircle(mo, mo.eflags & MFE_UNDERWATER and MT_SMALLBUBBLE or MT_DUST,
			8, 4*mo.scale, player.height/2, mo.scale, false, nil, true, player.drawangle)
		end
	end
	player.flarebouncex = nil
	player.flarebouncey = nil
	
	//Spin fire
	if player.pflags & PF_SPINNING and player.pflags & PF_THOKKED
		if P_IsObjectOnGround(mo)
			if mo.tics == 1 and not (mo.frame & FF_FRAMEMASK)
				S_StartSound(mo, sfx_flmove)
				
				if player.narlare and player.powers[pw_super]
					if player == displayplayer and not splitscreen
						P_StartQuake(8*mo.scale, 2)
					end
					if leveltime & 1
						S_StartSound(mo, sfx_flstp2)
					else
						S_StartSound(mo, sfx_flstp1)
					end
				end
			end
			
			if mo.state == S_PLAY_STND
				mo.state = S_PLAY_ROLL
			end
			player.panim = PA_ABILITY
			
			for i = 1, 2
				local fire = P_SpawnPlayerMissile(mo, MT_SPINFIRE, 0)
				if not (fire and fire.valid) then continue end
				
				fire.state = S_LASERFLAME3 + (leveltime % 3)
				fire.tics = max($, 1)
				fire.angle = R_PointToAngle2(0,0, mo.momx, mo.momy)
					
				local move = i & 1 and -1 or 1
				local radius = mo.radius*5/4*move
				if player.narlare and player.powers[pw_super]
					radius = $*2/3
				else
					radius = $/2
				end
				P_TryMove(fire,
				fire.x + P_ReturnThrustX(mo, fire.angle + ANGLE_90, radius),
				fire.y + P_ReturnThrustY(mo, fire.angle + ANGLE_90, radius),
				false)
				
				fire.color = player.powers[pw_super] and SUPERFIRE or SKINCOLOR_KETCHUP
				fire.colorized = true
				fire.blendmode = AST_ADD
				fire.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
				fire.spriteyoffset = $ - 8*FRACUNIT
				
				fire.flags = $ &~ MF_NOGRAVITY &~ MF_FIRE
				fire.radius = 24*fire.scale
			end
			
			//Spindash Switch
			if mo.subsector and mo.subsector.valid
			and mo.subsector.sector and mo.subsector.sector.valid
				for fof in mo.subsector.sector.ffloors()
					if fof.valid and mo.z + mo.momz <= fof.topheight
					and mo.z + mo.momz + mo.height >= fof.bottomheight
					and fof.master and fof.master.valid and fof.master.flags&ML_NOCLIMB
					and (fof.master.special == 153 or fof.master.special == 180
					or (fof.master.special >= 150 and fof.master.special <= 152)
					or (fof.master.special >= 190 and fof.master.special <= 195)
					or fof.master.special == 176 or fof.master.special == 177)
						player.pflags = $ | PF_STARTDASH
						A_ForceStop(mo, 1)
						return
					end
				end
			end
		end
		
		player.flarespinfire = true
		
	//Continuing after landing
	elseif player.flarespinfire
		if P_IsObjectOnGround(mo) and player.speed > 5*mo.scale
		and (not (not player.panim or player.panim >= PA_DASH) or mo.state == S_PLAY_ROLL)
			mo.state = S_PLAY_ROLL
			player.pflags = $ | PF_SPINNING | PF_THOKKED &~ PF_JUMPED &~ PF_NOJUMPDAMAGE
			
			if mo.eflags & (MFE_UNDERWATER|MFE_TOUCHWATER) // play a blunt sound
				S_StartSound(mo, sfx_s3k56)
				flarecircle(mo, MT_SMALLBUBBLE, 8, 4*mo.scale, player.height/2,
				mo.scale, false, nil, true, player.drawangle)
			else // create a fire pattern on the ground
				S_StartSound(mo, sfx_s3k47)
				P_ElementalFire(player, true)
			end
		elseif player.pflags & PF_SPINNING and not P_IsObjectOnGround(mo)
			player.pflags = $ | PF_THOKKED
		else
			player.flarespinfire = false
		end
	end
	
	//Landing after the jump ability
	if player.flarelanding
		player.flarelanding = $ + 1
		if ((not player.panim or player.panim >= PA_DASH)
		and mo.state != S_PLAY_GLIDE) or (not P_IsObjectOnGround(mo)
		and not (player.pflags & PF_JUMPED and player.pflags & PF_THOKKED))
			player.flarelanding = 0
			S_StopSoundByID(mo, sfx_s3k42)
			
			if mo.state == S_PLAY_GLIDE
				mo.state = S_PLAY_JUMP
				if not P_IsObjectOnGround(mo)
					player.pflags = $ | P_GetJumpFlags(player)
				end
			end
		elseif not P_IsObjectOnGround(mo)
			player.pflags = $ | PF_JUMPSTASIS
			mo.momz = $ + P_GetMobjGravity(mo) * (player.powers[pw_super] and 3 or 1)
		end
	end
	
	//Spinning
	if mo.state == S_PLAY_GLIDE
		//Ghost
		if player.flarelanding & 1
			local ghost = P_SpawnGhostMobj(mo)
			ghost.fuse = $/2
			
			if player.powers[pw_super]
				ghost.color = SUPERFIRE
				ghost.colorized = true
			end
		end
		player.powers[pw_justsprung] = max($, 2)
		
		if not player.powers[pw_super]
			//Fire
			local fire = P_SpawnMobjFromMobj(mo,0,0,4*FRACUNIT,MT_DUST)
			fire.state = S_SPINDUST_FIRE1
			fire.angle = player.drawangle
			
			local twod = mo.flags2 & MF2_TWOD or twodlevel
			or not (player.pflags & PF_DIRECTIONCHAR)
			if twod and player.flare2dspin != nil
				fire.angle = player.flare2dspin
			end
			
			fire.color = player.powers[pw_super] and SUPERFIRE or SKINCOLOR_CRIMSON
			fire.colorized = true
			fire.blendmode = AST_ADD
			
			local radius = mo.radius*5/4
			if player.narlare and player.powers[pw_super] then radius = $*2 end
			P_TeleportMove(fire,
			fire.x + P_ReturnThrustX(fire, fire.angle, radius),
			fire.y + P_ReturnThrustY(fire, fire.angle, radius),
			fire.z)
			
			fire.momx = mo.momx
			fire.momy = mo.momy
			fire.momz = mo.momz
			P_Thrust(fire, fire.angle, 2*fire.scale)
			
			//Spin
			if twod
				if player.flare2dspin == nil
					player.flare2dspin = player.drawangle
				else
					player.drawangle = player.flare2dspin
				end
			end
			if player.flarelanding <= 8
				player.drawangle = $ - ANGLE_45
			else
				player.drawangle = $ - ANGLE_22h
				if not (player.flarelanding % 16)
					S_StartSound(mo, sfx_s3k42)
				end
			end
			player.flare2dspin = player.drawangle
		end
	elseif player.flare2dspin != nil
		player.flare2dspin = nil
	end
	
	//State junk
	if (mo.state == S_PLAY_GLIDE or player.flarelanding) and P_IsObjectOnGround(mo)
		mo.state = S_PLAY_MELEE_LANDING
		mo.tics = TICRATE/3
		
		player.drawangle = mo.angle
		player.skidtime = 0
		
		mo.momx = $/2
		mo.momy = $/2
		
		S_StopSoundByID(mo, sfx_s3k42)
		S_StartSound(mo, sfx_s3k4c)
		
		if player.powers[pw_super]
			P_NukeEnemies(mo, mo, 512*FRACUNIT)
			S_StartSound(mo, sfx_flland)
			
			local fire = P_SpawnMobjFromMobj(player.mo,0,0,4*FRACUNIT,MT_THOK)
			fire.sprite = SPR_FLAM
			fire.frame = M
			fire.color = SUPERFIRE
			fire.colorized = true
			P_Thrust(fire, player.cmd.angleturn << 16, 32*fire.scale)
			fire.scale = $/2
			fire.scalespeed = fire.scale
			fire.destscale = $*16
			fire.flags2 = $|MF2_SPLAT|MF2_BOSSNOTRAP
			fire.blendmode = AST_ADD
			fire.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
			fire.fuse = 16
			
			local fa
			for i = 0, 15
				fa = (i*(ANGLE_180/8))
				fire = P_SpawnMobjFromMobj(player.mo,0,0,0,MT_THOK)
				fire.color = SUPERFIRE
				fire.colorized = true
				fire.state = S_LASERFLAME4
				fire.angle = InvAngle(fa)
				fire.momx = FixedMul(sin(fa),20*fire.scale)
				fire.momy = FixedMul(cos(fa),20*fire.scale)
				fire.spritexscale = FRACUNIT/3
				fire.flags2 = $|MF2_BOSSNOTRAP
				fire.blendmode = AST_ADD
				fire.renderflags = $ | RF_FULLBRIGHT | RF_NOCOLORMAPS
				fire.fuse = 16
			end
		end
			
		player.flarelanding = 0
		
	elseif mo.state == S_PLAY_MELEE_LANDING
		if P_IsObjectOnGround(mo) and player.speed
			P_SpawnSkidDust(player, mo.radius*5/4, not (leveltime % 3))
		end
		
		if mo.tics == 1 and player.cmd.buttons & BT_JUMP
			mo.tics = -1
		elseif mo.tics == -1 and not (player.cmd.buttons & BT_JUMP)
			mo.tics = 4
		end
		
	elseif mo.state == S_PLAY_BOUNCE_LANDING
		player.panim = PA_DASH
		if mo.tics == 1
		and (mo.frame & FF_FRAMEMASK) >= skins[player.skin].sprites[mo.sprite2].numframes - 1
			mo.state = S_PLAY_BOUNCE_LANDING
			player.panim = PA_DASH
			if mo.state == S_PLAY_BOUNCE
				mo.state = S_PLAY_ROLL
				player.flarespinfire = true
				player.pflags = $ | PF_THOKKED | PF_SPINNING
			end
		end
	
	elseif player.hugtime and mo.state == S_PLAY_HUG
		mo.frame = ($ &~ FF_FRAMEMASK) | ((player.hugtime/4) % skins[player.skin].sprites[mo.sprite2].numframes)
	end
end)

//Can't move when landing + fire frames
addHook("PreThinkFrame", do for player in players.iterate
	local mo = player.mo
	if not CantMove(player)
		if P_IsObjectOnGround(mo)
		and (mo.state == S_PLAY_GLIDE or mo.state == S_PLAY_MELEE_LANDING)
			player.powers[pw_noautobrake] = max($, 2)
			player.cmd.forwardmove = 0
			player.cmd.sidemove = 0
		end
	end
	
	if mo and mo.flaresavedstate
		if mo.skin == "flarethetrex" and mo.state == mo.flaresavedstate
			mo.sprite2 = mo.flaresavedsprite2
			mo.frame = mo.flaresavedframe
		end
		mo.flaresavedstate = nil
	end
end end)

//Fire frames part 2
addHook("PostThinkFrame", do for player in players.iterate
	local mo = player.mo
	if mo and mo.skin == "flarethetrex"
	and (player.weapondelay or (player.flarethrower and player.flarethrower.valid))
		local sprite2 = mo.sprite2 &~ FF_SPR2SUPER
		if FIRESPRITES[sprite2]
			mo.flaresavedstate = mo.state
			mo.flaresavedsprite2 = mo.sprite2
			mo.flaresavedframe = mo.frame
			
			mo.sprite2 = FIRESPRITES[sprite2] | ($ & FF_SPR2SUPER)
			mo.frame = mo.flaresavedframe &~ FF_FRAMEMASK
			mo.frame = $ + min(mo.flaresavedframe & FF_FRAMEMASK, skins[mo.skin].sprites[mo.sprite2].numframes - 1)
		end
	end
end end)

//Carrying (I'm going to die)
addHook("MobjMoveCollide", function(mo, mobj)
	if gametyperules & GTR_FRIENDLY and not mo.flarecarrying
	and mobj.player and not CantMove(mobj.player, true)
	and not (mobj.player.pflags & PF_SPINNING)
	and not (mobj.player.pflags & PF_THOKKED)
	and (P_IsObjectOnGround(mobj) or mobj.momz*P_MobjFlip(mobj) < 0)
	and mo.player and not CantMove(mo.player)
	and (mo.player.flarespinfire or mo.player.flarelanding
	or mo.state == S_PLAY_MELEE_LANDING or mo.state == S_PLAY_BOUNCE_LANDING)
	and mobj.z - mobj.scale <= mo.z + mo.height
	and mobj.z + mobj.height + mobj.scale >= mo.z
		S_StartSound(mobj, sfx_s3k4a)
		P_ResetPlayer(mobj.player)
		mobj.player.powers[pw_carry] = 15
		mobj.tracer = mo
		mobj.flarecarry = mo.player.drawangle
		mo.flarecarrying = 2
	end
end, MT_PLAYER)

freeslot("S_FLARE_EDGE")
states[S_FLARE_EDGE] = {SPR_PLAY, SPR2_EDGE, 12, nil, 0, 0, S_FLARE_EDGE}
addHook("PlayerThink", function(player)
	local mo = player.mo
	if mo and mo.flarecarry != nil
		local tracer = mo.tracer
		if not tracer or not tracer.player
		or not (not P_IsObjectOnGround(tracer)
		or tracer.player.flarespinfire or tracer.player.flarelanding
		or tracer.state == S_PLAY_MELEE_LANDING or tracer.state == S_PLAY_BOUNCE_LANDING)
		or player.powers[pw_carry] != 15
		or CantMove(player, true, true) or player.pflags & PF_JUMPED
		or R_PointToDist2(0,0, R_PointToDist2(mo.x, mo.y, tracer.x, tracer.y),
		abs((mo.z + mo.height/2) - (tracer.z + tracer.height/2))) > tracer.height*16
			if tracer then mo.tracer = nil end
			mo.flarecarry = nil
			
			local keepflags = player.pflags & (PF_JUMPED | PF_NOJUMPDAMAGE | PF_STARTJUMP)
			P_ResetPlayer(player)
			player.pflags = $ | keepflags
		else
			P_TryMove(mo, tracer.x, tracer.y, true)
			mo.z = tracer.z + ((mo.eflags & MFE_VERTICALFLIP) and -mo.height or tracer.height)
			
			if tracer.state == S_PLAY_BOUNCE_LANDING
				A_ForceStop(mo)
			else
				mo.momx = tracer.momx
				mo.momy = tracer.momy
				mo.momz = tracer.momz
			end
			
			local runspeed = FixedMul(mo.scale, player.runspeed)
			if tracer.state == S_PLAY_BOUNCE_LANDING
			or not tracer.player.flarespinfire
				if player.cmd.forwardmove or player.cmd.sidemove
				or tracer.state == S_PLAY_BOUNCE_LANDING
					if player.followitem == MT_TAILSOVERLAY
						if mo.state != S_PLAY_EDGE
							mo.state = S_PLAY_EDGE
						end
					elseif mo.state != S_FLARE_EDGE
						mo.state = S_FLARE_EDGE
					end
					if player.charflags & SF_FASTEDGE
						mo.tics = min($, 2)
					end
				elseif player.panim != PA_IDLE
					mo.state = S_PLAY_STND
				end
			elseif player.speed > runspeed
				mo.state = S_PLAY_RUN
			elseif player.panim != PA_WALK and player.speed <= runspeed
				mo.state = S_PLAY_WALK
			elseif player.panim == PA_WALK
				if player.speed > 12*mo.scale
					mo.tics = min($, 1)
				else
					mo.tics = min($, 2)
				end
			end
			
			if not player.cmd.forwardmove and not player.cmd.sidemove
			and tracer.player.flarespinfire and mo.state != S_FLARE_EDGE
			and player.panim != PA_IDLE and player.panim != PA_EDGE
				player.drawangle = tracer.player.drawangle - ANGLE_180
			elseif player.bot == 1 and tracer.player.flarelanding
				player.drawangle = tracer.player.drawangle
			else
				player.drawangle = $ + tracer.player.drawangle - mo.flarecarry
			end
			mo.flarecarry = tracer.player.drawangle
			
			if player.bot == 1
				player.pflags = $ | PF_FULLSTASIS
			end
			tracer.flarecarrying = 2
		end
	end
end)

addHook("PlayerCanDamage", function(player, mobj)
	if (player.mo and player.mo.flarecarry)
	or (player.narlare and player.powers[pw_super])
		return true
	end
end)